from django.conf import settings
from django.db.models.lookups import Lookup

from pgcrypto import DIGEST_SQL, HMAC_SQL

PGCRYPTO_KEY = settings.PGCRYPTO_KEY


class HashLookupBase(Lookup):
    """Lookup to filter hashed values.

    `HashLookup` is hashing the value on the right hand side with
    the function specified in `encrypt_sql`.
    """
    lookup_name = 'hash_of'

    def as_sql(self, qn, connection):
        """Responsible for creating the lookup with the digest SQL.

        Modify the right hand side expression to compare the value passed
        to a hash.
        """
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params

        rhs = self.encrypt_sql % rhs
        return ('{}::bytea = {}'.format(lhs, rhs)), params


class DigestLookup(HashLookupBase):
    """Digest lookup producing a hash.

    `encrypt_sql` uses pgcrypto 'digest' function to create a hash based version
    of the field's value.
    """
    encrypt_sql = DIGEST_SQL


class HMACLookup(HashLookupBase):
    """HMAC lookup producing a hash.

    `encrypt_sql` uses pgcrypto 'hmac' function to create a hash based version
    the field's value.
    """
    encrypt_sql = HMAC_SQL


class DateLookupBase(Lookup):
    """Base class for Date lookups."""
    lookup_name = None  # Set in subclasses
    operator = None  # Set in subclasses

    def __init__(self, lhs, rhs):
        """Implementing an abstract class."""
        super(DateLookupBase, self).__init__(lhs, rhs)  # pragma: no cover

    def as_sql(self, qn, connection):
        """Build SQL with decryption and casting."""
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        return "cast(pgp_sym_decrypt(%s, '%s') as DATE) %s %s" % (
            lhs, PGCRYPTO_KEY, self.operator, rhs
        ), params


class DateGT(DateLookupBase):
    """Lookup for Date."""
    lookup_name = 'gt'
    operator = '>'


class DateGTE(DateLookupBase):
    """Lookup for Date."""
    lookup_name = 'gte'
    operator = '>='


class DateLT(DateLookupBase):
    """Lookup for Date."""
    lookup_name = 'lt'
    operator = '<'


class DateLTE(DateLookupBase):
    """Lookup for Date."""
    lookup_name = 'lte'
    operator = '<='


class DateEXACT(DateLookupBase):
    """Lookup for Date."""
    lookup_name = 'exact'
    operator = '='


class DateTimeLookupBase(Lookup):
    """Base class for DateTime lookups."""
    lookup_name = None  # Set in subclasses
    operator = None  # Set in subclasses

    def as_sql(self, qn, connection):
        """Build SQL with decryption and casting."""
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        return "cast(pgp_sym_decrypt(%s, '%s') as TIMESTAMP) %s %s" % (
            lhs, PGCRYPTO_KEY, self.operator, rhs
        ), params


class DateTimeGT(DateTimeLookupBase):
    """Lookup for DateTime."""
    lookup_name = 'gt'
    operator = '>'


class DateTimeGTE(DateTimeLookupBase):
    """Lookup for DateTime."""
    lookup_name = 'gte'
    operator = '>='


class DateTimeLT(DateTimeLookupBase):
    """Lookup for DateTime."""
    lookup_name = 'lt'
    operator = '<'


class DateTimeLTE(DateTimeLookupBase):
    """Lookup for DateTime."""
    lookup_name = 'lte'
    operator = '<='


class DateTimeEXACT(DateTimeLookupBase):
    """Lookup for DateTime."""
    lookup_name = 'exact'
    operator = '='



from django.db import models

from pgcrypto import (
    DIGEST_SQL,
    HMAC_SQL,
    INTEGER_PGP_PUB_ENCRYPT_SQL,
    INTEGER_PGP_SYM_ENCRYPT_SQL,
    PGP_PUB_ENCRYPT_SQL,
    PGP_SYM_ENCRYPT_SQL,
)
from pgcrypto.lookups import (
    DateEXACT,
    DateGT,
    DateGTE,
    DateLT,
    DateLTE,
    DateTimeEXACT,
    DateTimeGT,
    DateTimeGTE,
    DateTimeLT,
    DateTimeLTE,
    DigestLookup,
    HMACLookup,
)
from pgcrypto.mixins import (
    DatePGPSymmetricKeyFieldMixin,
    DateTimePGPSymmetricKeyFieldMixin,
    EmailPGPPublicKeyFieldMixin,
    EmailPGPSymmetricKeyFieldMixin,
    HashMixin,
    PGPPublicKeyFieldMixin,
    PGPSymmetricKeyFieldMixin,
)


class TextDigestField(HashMixin, models.TextField):
    """Text digest field for postgres."""
    encrypt_sql = DIGEST_SQL


TextDigestField.register_lookup(DigestLookup)


class TextHMACField(HashMixin, models.TextField):
    """Text HMAC field for postgres."""
    encrypt_sql = HMAC_SQL


TextHMACField.register_lookup(HMACLookup)


class EmailPGPPublicKeyField(EmailPGPPublicKeyFieldMixin, models.EmailField):
    """Email PGP public key encrypted field."""
    encrypt_sql = PGP_PUB_ENCRYPT_SQL


class IntegerPGPPublicKeyField(PGPPublicKeyFieldMixin, models.IntegerField):
    """Integer PGP public key encrypted field."""
    encrypt_sql = INTEGER_PGP_PUB_ENCRYPT_SQL


class TextPGPPublicKeyField(PGPPublicKeyFieldMixin, models.TextField):
    """Text PGP public key encrypted field."""
    encrypt_sql = PGP_PUB_ENCRYPT_SQL


class EmailPGPSymmetricKeyField(EmailPGPSymmetricKeyFieldMixin, models.EmailField):
    """Email PGP symmetric key encrypted field."""
    encrypt_sql = PGP_SYM_ENCRYPT_SQL


class IntegerPGPSymmetricKeyField(PGPSymmetricKeyFieldMixin, models.IntegerField):
    """Integer PGP symmetric key encrypted field."""
    encrypt_sql = INTEGER_PGP_SYM_ENCRYPT_SQL


class TextPGPSymmetricKeyField(PGPSymmetricKeyFieldMixin, models.TextField):
    """Text PGP symmetric key encrypted field for postgres."""
    encrypt_sql = PGP_SYM_ENCRYPT_SQL


class DatePGPSymmetricKeyField(DatePGPSymmetricKeyFieldMixin, models.TextField):
    """Date PGP symmetric key encrypted field for postgres."""
    encrypt_sql = PGP_SYM_ENCRYPT_SQL
    cast_sql = 'cast(%s as DATE)'


DatePGPSymmetricKeyField.register_lookup(DateEXACT)
DatePGPSymmetricKeyField.register_lookup(DateGT)
DatePGPSymmetricKeyField.register_lookup(DateGTE)
DatePGPSymmetricKeyField.register_lookup(DateLT)
DatePGPSymmetricKeyField.register_lookup(DateLTE)


class DateTimePGPSymmetricKeyField(DateTimePGPSymmetricKeyFieldMixin, models.TextField):
    """DateTime PGP symmetric key encrypted field for postgres."""
    encrypt_sql = PGP_SYM_ENCRYPT_SQL
    cast_sql = 'cast(%s as TIMESTAMP)'


DateTimePGPSymmetricKeyField.register_lookup(DateTimeEXACT)
DateTimePGPSymmetricKeyField.register_lookup(DateTimeGT)
DateTimePGPSymmetricKeyField.register_lookup(DateTimeGTE)
DateTimePGPSymmetricKeyField.register_lookup(DateTimeLT)
DateTimePGPSymmetricKeyField.register_lookup(DateTimeLTE)



from django.conf import settings
from django.db import models


class PGPManager(models.Manager):
    use_for_related_fields = True
    use_in_migrations = True

    @staticmethod
    def _get_pgp_symmetric_decrypt_sql(field):
        """Decrypt sql for symmetric fields using the cast sql if required."""
        sql = """pgp_sym_decrypt("{0}"."{1}", '{2}')"""
        if hasattr(field, 'cast_sql'):
            sql = field.cast_sql % sql

        return sql.format(
            field.model._meta.db_table,
            field.name,
            settings.PGCRYPTO_KEY,
        )

    @staticmethod
    def _get_pgp_public_key_decrypt_sql(field):
        """Decrypt sql for public key fields using the cast sql if required."""
        sql = """pgp_pub_decrypt("{0}"."{1}", dearmor('{2}'))"""

        if hasattr(field, 'cast_sql'):
            sql = field.cast_sql % sql

        return sql.format(
            field.model._meta.db_table,
            field.name,
            settings.PRIVATE_PGP_KEY,
        )

    def get_queryset(self, *args, **kwargs):
        """Decryption in queryset through meta programming."""
        # importing here otherwise there's a circular reference issue
        from pgcrypto.mixins import PGPSymmetricKeyFieldMixin, PGPPublicKeyFieldMixin

        skip_decrypt = kwargs.pop('skip_decrypt', None)

        qs = super().get_queryset(*args, **kwargs)

        # The Django admin skips this process because it's extremely slow
        if not skip_decrypt:
            select_sql = {}
            encrypted_fields = []

            for field in self.model._meta.get_fields():
                if isinstance(field, PGPSymmetricKeyFieldMixin):
                    select_sql[field.name] = self._get_pgp_symmetric_decrypt_sql(field)
                    encrypted_fields.append(field.name)
                elif isinstance(field, PGPPublicKeyFieldMixin):
                    select_sql[field.name] = self._get_pgp_public_key_decrypt_sql(field)
                    encrypted_fields.append(field.name)

            # Django queryset.extra() is used here to add decryption sql to query.
            qs = qs.defer(
                *encrypted_fields
            ).extra(
                select=select_sql
            )

        return qs



from django import forms


class DateField(forms.DateField):
    """
    Form class for DateField.

    Removes max_length since the backend model field is a text field.
    """

    def __init__(self, input_formats=None, *args, **kwargs):
        """Init that pops off the max_length attribute."""
        kwargs.pop('max_length', None)
        super().__init__(input_formats, *args, **kwargs)


class DateTimeField(forms.DateTimeField):
    """
    Form class for DateTimeField.

    Remove max_length since the backend model field is a text field.
    """

    def __init__(self, input_formats=None, *args, **kwargs):
        """Init that pops off the max_length attribute."""
        kwargs.pop('max_length', None)
        super().__init__(input_formats, *args, **kwargs)



class PGPAdmin(object):

    def get_queryset(self, request):
        """Skip any auto decryption when ORM calls are from the admin."""
        return self.model.objects.get_queryset(**{'skip_decrypt': True})
